/* im.c - image routines */

#include <assert.h>
#include <math.h>
#include "im.h"
#include "xmalloc.h"


/* create a new image */
im_t *
im_new(int w, int h, int channels)
{
	im_t *im = xmalloc(sizeof(im_t));

	im->w = w;
	im->h = h;
	im->channels = channels;
	im->data = xmalloc(w*h*channels*sizeof(int));

	return im;
}

/* free the memory allocated for an image */
void
im_free(im_t *im)
{
	xfree(im->data);
	xfree(im);
}

/* create a new image for rotozooming */
im_t *
im_new_rotozoom(im_t *orig)
{
	int side, w, h, i, k, x, y;
	im_t *im;
	int *p, *q;

	w = orig->w;
	h = orig->h;

	/* first calculate the side length of a bounding box that is big enough to
	   contain each rotation of the original rectangle. this is called the
	   sprite square and we will map it to the screen when drawing.

	   then calculate another bounding box that is big enough to contain each
	   rotation of the sprite square. this is the physical bitmap we store in
	   memory, and it optimizes the rendering loop as we don't have to check
	   whether we are inside the bitmap. */
	side = (int)sqrt(w*w+h*h);
	side = (int)(sqrt(2)*side + 5); /* plus 5 pixel error margin */

	im = im_new(side,side,1);
	p = im->data; i = side*side;
	while (i--) *p++ = 0;

	/* copy the original bitmap to the center of the newly allocated image */
	y = side/2 - h/2;
	x = side/2 - w/2;
	q = orig->data;

	for (i=0; i<h; i++,y++)
	{
		p = im->data + y*side+x;
		k=w; while (k--) *p++ = *q++;
	}

	return im;
}

/* draw the image rotated and zoomed to the destination buffer */
void
im_rotozoom(im_t *im, im_t *buf, double a, double zx, double zy, double px, double py, double c)
{
	double sa = sin(a+M_PI);
	double ca = cos(a+M_PI);
	int side = (int)((double)im->w/sqrt(2));
	vec_t xb,yb,co,x;
	int cx,cy,ex,ey,i,k,w;
	int *dst;

	(void)c;

	/* calculate base vectors */
	xb[0]= ca;
	xb[1]=-sa;
	yb[0]=-sa;
	yb[1]=-ca;

	/* calculate corner coordinates in the bitmap space */
	co[0]=-side/2*xb[0] - side/2*yb[0] + im->w/2;
	co[1]=-side/2*xb[1] - side/2*yb[1] + im->w/2;

	/* scale the base vectors */
	vec_mul(xb,1/zx);
	vec_mul(yb,1/zy);

	/* calculate the coordinates of upper left corner */
	cx = px - side*zx/2 + buf->w/2;
	cy = py - side*zy/2 + buf->h/2;
	ex = side*zx;
	ey = side*zy;

	/* clip clap */
	if (cx+side*zx > buf->w) ex -= cx+side*zx - buf->w;
	if (cx<0)
	{
		ex += cx;
		co[0] += -cx*xb[0];
		co[1] += -cx*xb[1];
		cx = 0;
	}
	if (cy+side*zy > buf->h) ey -= cy+side*zy - buf->h;
	if (cy<0)
	{
		ey += cy;
		co[0] += -cy*yb[0];
		co[1] += -cy*yb[1];
		cy = 0;
	}

	/* render the picture */
	w = im->w;
	for (i=0; i<ey; i++)
	{
		dst = buf->data + cy*buf->w+cx;
		vec_mov(x,co);

		for (k=0; k<ex; k++)
		{
			*dst++ += im->data[(int)x[1]*w+(int)x[0]];
			vec_add(x,xb);
		}
		vec_add(co,yb);
		cy++;
	}
}

/* crossfade between two images */
void
im_crossfade(im_t *a, im_t *b, double t)
{
	int i = a->w*a->h*a->channels;
	int *ap = a->data;
	int *bp = b->data;
	int u = (int)(65536.0*t);
	int v = (int)(65536.0*(1.0-t));

	while (i--)
	{
		*ap = (v**ap + u**bp)>>16;
		ap++,bp++;
	}
}

/* find histogram of the given image.
   NOTE: expect pixel values to be within [0,255]. */
double *
histogram(im_t *im, double *hist)
{
	double d;
	int *p, size;

	assert(im->channels == 1);

	if (hist == NULL)
		hist = xmalloc(256*sizeof(double));

	p = im->data;
	size = im->w*im->h;
	d = 1.0/(double)size;

	while (size--)
		hist[*p++] += d;

	return hist;
}

/* histogram equalization */
im_t *
histogram_equalize(im_t *s, im_t *d)
{
	double hist[256]={0}, sum;
	int lut[256], i, size, *sp, *dp;

	if (d == NULL)
		d = im_new(s->w, s->h, s->channels);

	assert(s->channels == 1 && d->channels == 1);
	assert(s->w*s->h <= d->w*d->h);

	histogram(s, hist);

	sum = 0;
	for (i = 0; i < 256; i++)
	{
		sum += hist[i];
		lut[i] = (int)(sum*200.0);
	}

	sp = s->data;
	dp = d->data;
	size = s->w*s->h;
	while (size--)
		*dp++ = lut[*sp++];

	return d;
}
