/* script.c - simple animation engine */

#include <math.h>
#include "main.h"
#include "script.h"
#include "xmalloc.h"
#include "util.h"
#include "input.h"

/* random numbers */
#define SEQ_SIZE 256

void seq_advance(int);
void seq_set_ticks(int);
void seq_put(int);
int seq_get(int);

int seq[SEQ_SIZE],seq_i=0;

/* data and variables */
im_t *pics[SCRIPT_MAX_PICS];
int pics_n = 0;
static int ticks = 0;
static list_t *events = NULL;
static im_t *back,*front,*tmpbuf;
static int smooth[256],smooth3[100];
extern double suodatin[INPUT_NUM][INPUT_LEN],displace[INPUT_NUM][INPUT_LEN];
extern int metaframes;


void
seq_put(int x)
{
	seq[seq_i++] = x;
	seq_i = seq_i%SEQ_SIZE;
}

int
seq_get(int i)
{
	int k,diff;

	diff = i-ticks;
	k = (seq_i+diff)%SEQ_SIZE;
	if (k<0) k=SEQ_SIZE-k;
	return seq[k];
}

void
seq_advance(int n)
{
	if (n<0)
	{
		mysrand(0);
		n = ticks;
	}

	if (n<0) return;

	while (n--)
		seq_put(myrand());
}

void
seq_set_ticks(int t)
{
	seq_i=0;
	mysrand(0);
	seq_advance(t);
}

void
script_init(void)
{
	back = im_new(BUF_WIDTH,BUF_HEIGHT,1);
	front = im_new(BUF_WIDTH,BUF_HEIGHT,1);
	tmpbuf = im_new(BUF_WIDTH,BUF_HEIGHT,1);
	filter_build_gaussian1d(smooth,3.0,5);
	filter_build_gaussian1d(smooth3,3.0,3);
	seq_set_ticks(0);
}

script_t *
script_add(void)
{
	script_t *e = xmalloc(sizeof(script_t));

	e->p = 0;
	e->t = ticks;
	e->d = 50;
	e->x = e->y = 100;
	e->a = 0.0;
	e->u = 1.0;
	e->v = 1.0;
	e->f = 1;

	events = list_add_head(events, e);
	return e;
}

void
script_remove(void)
{
	if (!events) return;
	xfree(events->key);
	events = list_delete(events);
}

void
script_update(void)
{
	script_t *event;
	list_t *l, *best;
	int best_time;

	if (!events) return;

	best_time = 1000000;
	best = events;

	l = events;
	do
	{
		l = l->next;
		event = l->key;
		if (ticks >= event->t && ticks <= event->t+event->d && abs(event->t-ticks) < best_time)
		{
			best_time = abs(event->t-ticks);
			best = l;
		}
	}
	while (l != events);

	events = best;
}

script_t *
script_rotate(void)
{
	script_t *event;
	list_t *l;

	if (!events) return NULL;

	l = events;
	do
	{
		l = l->next;
		event = l->key;
		if (ticks >= event->t && ticks <= event->t+event->d)
			break;
	}
	while (l != events);

	events = l;
	return script_get();
}

script_t *script_get(void) {
	if (!events) return NULL; return events->key;
}

int script_advance(int t) {
	ticks += t; seq_advance(t); return ticks;
}
int script_set_ticks(int t) {
	seq_set_ticks(t); ticks = t; return ticks;
}

void
script_draw(im_t *buf)
{
	list_t *l;
	script_t *event;
	int i,nb,nf;

	for (i=0; i<BUF_WIDTH*BUF_HEIGHT; i++)
	{
		back->data[i] = 0;
		front->data[i] = 0;
	}

	if (!events)
	{
		for (i=0; i<buf->w*buf->h; i++)
			buf->data[i]=0;
		return;
	}

	nb = nf = 0;
	l = events;
	do
	{
		event = l->key;
		if (event->t <= ticks && ticks <= event->t+event->d)
		{
			double t = (double)(ticks-event->t)/(double)event->d;

			if (!event->f)
			{
				im_rotozoom(pics[event->p], back, event->a, event->u, event->v, event->x,event->y, sin(t*M_PI/2.0+M_PI/4.0));
				nb++;
			}
			else
			{
//              int x = (int)(input_cur(0)-320);
//              int y = (int)(input_cur(1)-240);
				int x = 0;
				int y = 0;
				im_rotozoom(pics[event->p], front, event->a, event->u, event->v, event->x+x,event->y+y, sin(t*M_PI/2.0+M_PI/4.0));
				nf++;
			}
		}
		l = l->next;
	}
	while (l != events);

	if (nb>0)
	{
		buffer_mul(back,1.0/(double)nb);
		histogram_equalize(back,back);
	}
	buffer_mul(front,1.75/(double)(nf ? nf : 1.75));
	buffer_clip(front);
	buffer_crossfade(back,front,buf,0.95);
}

void
buffer_add(im_t *a, im_t *b, double c)
{
	int *ap = a->data;
	int *bp = b->data;
	int n = a->w*a->h;
	int cc = (int)(65536.0*c);

	while (n--)
		*ap++ += (cc**bp++)>>16;
}

void
buffer_mul(im_t *a, double c)
{
	int *ap = a->data;
	int n = a->w*a->h;
	int cc = (int)(65536.0*c);

	while (n--)
	{
		*ap = (cc**ap)>>16;
		ap++;
	}
}

void
buffer_crossfade(im_t *a, im_t *b, im_t *c, double d)
{
	int i = a->w*a->h;
	int *ap = a->data;
	int *bp = b->data;
	int *cp = c->data;
	int u = (int)(65536.0*d);
	int v = (int)(65536.0*(1.0-d));

	while (i--)
	{
		if (*bp)
			*cp = (v**ap + u**bp)>>16;
		else
			*cp = *ap;
		ap++,bp++,cp++;
	}
}

void
buffer_invert(im_t *a)
{
	int *ap = a->data;
	int n = a->w*a->h;

	while (n--)
	{
		*ap = 255-*ap;
		ap++;
	}
}

void
buffer_clip(im_t *a)
{
	int *ap = a->data;
	int n = a->w*a->h;

	while (n--)
	{
		if (*ap<0) *ap=0;
		if (*ap>255) *ap=255;
		ap++;
	}
}

void
script_render(im_t *src, uint8 *fp)
{
	int i,k,c,n;
	int *q;

//  c = (int)(65536.0*(1.5-input_cur(0)/320.0));
	c = (int)(65536.0*(1.5-suodatin[0][metaframes]/320.0));
//  c = 65536.0;

	q = src->data;
	n = src->w*src->h;
	if (c>=0)
	{
		while (n--)
		{
			*q = (c**q)>>16;
			if (*q>255) *q=255;
			q++;
		}
	}
	else
	{
		while (n--)
		{
			*q = ((c**q)>>16)&0xff;
			q++;
		}
	}

	n = (src->h-5)*src->w;
	q = src->data+n;
	while (n--)
		*q++ = 0;

//  if (input_cur(2)==0.0)
	if (suodatin[2][metaframes]==0.0)
		im_convolve_separable_5x5(src,src,smooth,smooth);
	else
		im_convolve_separable_3x3(src,src,smooth3,smooth3);

	for (i=0; i<BUF_HEIGHT*2; i++)
	{
		q = src->data+(i/2)*BUF_WIDTH;
		for (k=0; k<BUF_WIDTH; k++)
		{
			*fp++ = q[k];
			*fp++ = q[k];
			*fp++ = q[k];
			*fp++ = q[k];
			*fp++ = q[k];
			*fp++ = q[k];
		}
	}
}

void
script_load(char *path)
{
	script_t *e;
	FILE *f;

	f = fopen(path, "r");

	while (1)
	{
		e=script_add();
		if (fscanf(f, "%d %d %d %d %d %lf %lf %lf %d\n", &e->p,&e->t,&e->d,&e->x,&e->y,&e->a,&e->u,&e->v,&e->f) != 9)
			break;
	}
	script_remove();

	fclose(f);
}

void
script_save(char *path)
{
	list_t *l;
	script_t *e;
	FILE *f;

	if (!events) return;

	f = fopen(path, "w");

	l = events;
	do
	{
		e = l->key;
		fprintf(f, "%d %d %d %d %d %f %f %f %d\n", e->p,e->t,e->d,e->x,e->y,e->a,e->u,e->v,e->f);
		l = l->next;
	}
	while (l != events);

	fclose(f);
}
